@@ -2,10 +2,14 @@ import atexit
 
 from copy      import deepcopy
 from sys       import getrefcount
+from os        import close
 from os        import remove
+from os        import rmdir
+from os.path   import isfile
 from operator  import mul
 from itertools import izip
 from itertools import product as itertools_product
+from tempfile  import mkstemp
 
 from numpy import array       as numpy_array
 from numpy import bool_       as numpy_bool_
@@ -27,6 +31,7 @@ from numpy.ma.core import MaskedConstant as numpy_ma_core_MaskedConstant
 from ..units     import Units
 from ..functions import get_subspace, FREE_MEMORY, FM_THRESHOLD
 from ..functions import inspect as cf_inspect
+from ..constants import CONSTANTS
 
 from .filearray import FileArray, _TempFileArray #, SharedMemoryArray, _shared_memory_array
 
@@ -35,14 +40,32 @@ _Units_conform = Units.conform
 _dtype_object = numpy_dtype(object)
 
 # --------------------------------------------------------------------
-# Set of partitions' temporary files
+# Dictionary of partitions' temporary files containing the full path
+# of the directory containing tuples of the temporary file and its
+# lock files, the full path of the lock file on this PE and a set of
+# the full path of the lock files on other PEs.
 #
 # For example:
 # >>> _temporary_files
-# set(['/tmp/cf_array_B8SSw2.npy',
-#      '/tmp/cf_array_iRekAW.npy'])
+# {'/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy': ('/tmp/cf_array_7hoQk5',
+#  '/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_i043rk',
+#  set('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_cvUcib')),
+#  '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
+#  '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
+#  set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
 # --------------------------------------------------------------------
-_temporary_files = set()
+_temporary_files = {}
+
+def _lock_files_present(lock_files):
+    lock_files_present = False
+    for filename in lock_files:
+        if isfile(filename):
+            lock_files_present = True
+            break
+        #--- End: if
+    #--- End: for
+    return lock_files_present
+#--- End: def
 
 def _remove_temporary_files(filename=None):
     '''Remove temporary partition files from disk.
@@ -75,36 +98,69 @@ not recommended to be used as a general tidy-up function.
 :Examples 2:
 
 >>> _temporary_files
-set(['/tmp/cf_array_B8SSw2.npy',
-     '/tmp/cf_array_G756ks.npy',
-     '/tmp/cf_array_iRekAW.npy'])
->>> _remove_temporary_files('/tmp/cf_array_G756ks.npy')
+{'/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy': ('/tmp/cf_array_7hoQk5',
+ '/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_i043rk',
+ set('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy_cvUcib')),
+ '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
+ '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
+ set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
+>>> _remove_temporary_files('/tmp/cf_array_7hoQk5/cf_array_RmC7NJ.npy')
 >>> _temporary_files
-set(['/tmp/cf_array_B8SSw2.npy',
-     '/tmp/cf_array_iRekAW.npy'])
+{'/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy': ('/tmp/cf_array_CrbJVf',
+ '/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_n4BKBF',
+ set('/tmp/cf_array_CrbJVf/cf_array_INZQQc.npy_xQ5LVo'))}
 >>> _remove_temporary_files()
 >>> _temporary_files
-set()
+{}
 
     '''
     if filename is not None:
         if filename in _temporary_files:
-            # Remove the given temporary file
+            # If this condition is not met then probably
+            # _remove_temporary_files() has already been run at
+            # exit
+            dirname, _lock_file, _other_lock_files = _temporary_files[filename]
             try:
-                remove(filename)
+                remove(_lock_file)
             except OSError:
                 pass
-            _temporary_files.remove(filename)
-        #--- End: if
+            #--- End: if
+
+            # Only remove the temporary file if it is not being
+            # used by any other ranks
+            if not _lock_files_present(_other_lock_files):
+                # Remove the given temporary file
+                try:
+                    remove(filename)
+                    rmdir(dirname)
+                except OSError:
+                    pass
+                del _temporary_files[filename]
+            #--- End: if
         return
     #--- End: if
 
-    # Still here? Then remove all temporary files
+    # Still here? Then remove all temporary files and lock files
     for filename in _temporary_files:
         try:
             remove(filename)
         except OSError:
             pass
+        _dirname, _lock_file, _other_lock_files = _temporary_files[filename]
+        try:
+            remove(_lock_file)
+        except OSError:
+            pass
+        for lockfile in _other_lock_files:
+            try:
+                remove(lockfile)
+            except OSError:
+                pass
+        #--- End: for
+        try:
+            rmdir(dirname)
+        except:
+            pass
     #--- End: for
 
     _temporary_files.clear()
@@ -260,8 +316,9 @@ by any other partition then the file is closed.
         _partition_file = getattr(subarray, '_partition_file', None)
         if _partition_file is not None:
             # This partition contains a temporary file which is not
-            # referenced by any other partition, so remove the file
-            # from disk.
+            # referenced by any other partition on this process, so if
+            # there are no lock files present remove the file from
+            # disk.
             _remove_temporary_files(_partition_file)
 
         else:
@@ -912,7 +969,8 @@ values of the partition's `!_original` attribute and on the
                             print '    1.2.2.3 del masked'
                         del self.masked       
         else:
-            if _debug: print 'Partition.close: parallel'
+            if _debug:
+                print 'Partition.close: parallel'
             # --------------------------------------------------------
             # PARALLEL
             # --------------------------------------------------------
@@ -1772,8 +1830,13 @@ memory, but this is not checked.
         except:
             return False
 
+        fd, _lock_file = mkstemp(prefix=tfa._partition_file + '_',
+                                 dir=tfa._partition_dir)
+        close(fd)
+
         self.subarray = tfa
-        _temporary_files.add(tfa._partition_file)
+        _temporary_files[tfa._partition_file] = (tfa._partition_dir,
+                                                 _lock_file, set())
 
         if reopen:
             # Re-open the partition
@@ -1878,18 +1941,47 @@ place.
         self.__dict__ = other.__dict__.copy()
     #--- End: def
 
-#    def set_from_child_process(self, other):
-#        '''
-#        '''
-#        if other.in_shared_memory and self.in_shared_memory:
-#            # If BOTH subarrays are in shared memory
-#            _shared_memory_array[id(other.subarray)] = _shared_memory_array[id(self.subarray)] 
-#            
-#        if other.in_temporary_file:
-#            # The other subarray is in a temporary file
-#            _temporary_files.add(other._subarray._partition_file)
-#            
-#        self.__dict__ = other.__dict__
-#    #--- End: def
+    # def set_from_child_process(self, other):
+    #     '''
+    #     '''
+    #     if other.in_shared_memory and self.in_shared_memory:
+    #         # If BOTH subarrays are in shared memory
+    #         _shared_memory_array[id(other.subarray)] = _shared_memory_array[id(self.subarray)] 
+            
+    #     if other.in_temporary_file:
+    #         # The other subarray is in a temporary file
+    #         _temporary_files.add(other._subarray._partition_file)
+            
+    #     self.__dict__ = other.__dict__
+    # #--- End: def
     
+    def _register_temporary_file(self):
+        # Register a temporary file on this rank that has been created
+        # on another rank
+        _partition_file = self._subarray._partition_file
+        _partition_dir = self._subarray._partition_dir
+        if not _partition_file in _temporary_files:
+            fd, _lock_file = mkstemp(prefix=_partition_file + '_',
+                                     dir=_partition_dir)
+            close(fd)
+            _temporary_files[_partition_file] = (_partition_dir,
+                                                 _lock_file, set())
+        else:
+            _, _lock_file, _ = _temporary_files[_partition_file]
+        #--- End: if
+        return _lock_file
+    #--- End: def
+
+    def _update_lock_files(self, lock_files):
+        # Add the lock files listed in lock_files to the list of lock
+        # files managed by other ranks
+        _, _lock_file, _other_lock_files = _temporary_files[self._subarray._partition_file]
+        _other_lock_files.update(set(lock_files))
+        if _lock_file in _other_lock_files:
+            # If the lock file managed by this rank is in the list of
+            # lock files managed by other ranks, remove it from there
+            _other_lock_files.remove(_lock_file)
+        #--- End: if
+    #--- End: def
+
 #--- End: class
